#include "header.h"
#define FLAVOR 1
#include "common.h"

/**
 * Basic checks for gather
 */

void t(void*_){
    switch (thread_myid()){
        case  2: [[fallthrough]];
        case  5: [[fallthrough]];
        case  6: wait_state(1); break;
        case  3: [[fallthrough]];
        case  4: [[fallthrough]];
        case  8: wait_state(2); break;
        case  7: [[fallthrough]];
        case  9: [[fallthrough]];
        case 10: [[fallthrough]];
        case 13: wait_state(3); break;
        case 11: [[fallthrough]];
        case 12: wait_state(4); break;
        default: assert(0);
    }
    
    thread_send(1, thread_msg_make(thread_myid()));
}

void g(void*_){
    wait_state(1);
    
    int ids[5]={2,5,6};
    thread_msg* msgs[5];
    int rv = thread_gather(ids, 3, msgs);
    assertVerbose(rv==3, "Should gather from all 3 threads, got %d (%p %p %p)", rv, msgs[0], msgs[1], msgs[2]);
    assert(msgs[0]->num==2); thread_msg_dec_use_count(msgs[0]);
    assert(msgs[1]->num==5); thread_msg_dec_use_count(msgs[1]);
    assert(msgs[2]->num==6); thread_msg_dec_use_count(msgs[2]);
    
    adv_state(2);
    
    ids[0]=8;
    ids[1]=3;
    ids[2]=2;
    ids[3]=4;
    rv = thread_gather(ids, 4, msgs);
    assertVerbose(rv==3, "Should gather from 3 threads out of 4, one is dead, got %d (%p %p %p %p)", rv, msgs[0], msgs[1], msgs[2], msgs[3]);
    assert(msgs[0]->num==8); thread_msg_dec_use_count(msgs[0]);
    assert(msgs[1]->num==3); thread_msg_dec_use_count(msgs[1]);
    assert(msgs[2]==NULL);
    assert(msgs[3]->num==4); thread_msg_dec_use_count(msgs[3]);
    
    adv_state(3);
    
    usleep(125000); // make sure these threads queue their messages
    
    adv_state(4);
    
    usleep(125000); // make these threads start waiting
    
    ids[0]=99;
    ids[1]=7;
    ids[2]=10;
    rv = thread_gather(ids, 3, msgs);
    assertVerbose(rv==2, "Should gather from 2 threads out of 3, one is invalid, got %d (%p %p %p)", rv, msgs[0], msgs[1], msgs[2]);
    assert(msgs[0]==NULL);
    assert(msgs[1]->num==7);  thread_msg_dec_use_count(msgs[1]);
    assert(msgs[2]->num==10); thread_msg_dec_use_count(msgs[2]);
    
    ids[0]=9;
    ids[1]=11;
    rv = thread_gather(ids, 2, msgs);
    assertVerbose(rv==2, "Should gather from all 2 threads that started waiting in send before previous gather, got %d (%p %p)", rv, msgs[0], msgs[1]);
    assert(msgs[0]->num==9);  thread_msg_dec_use_count(msgs[0]);
    assert(msgs[1]->num==11); thread_msg_dec_use_count(msgs[1]);
    
    adv_state(5);
}

int main(){
    printf("08_gather_basic: ");
    fflush(0);
    thread_init();
 
    int id = thread_make(g, NULL);
    assertVerbose(id==1, "bad thread IDs");
    
    for(int i = 2 ; i <= 13; ++i){
        int id2 = thread_make(t, NULL);
        assertVerbose(id2==i, "bad thread IDs");
    }
    
    adv_state(1);
    
    wait_state(5);
    
    usleep(50000);
    
    puts("ok");
    return 0;
}
