#!/bin/bash

if (($#))
then
    rm -f libheader.so
    gcc --std=gnu17 -DTHREAD_MSGQUEUE_SIZE=4 -g -fPIC -shared "$1" -I. -o libheader.so || { echo "DOES NOT COMPILE"; exit; }
else
    echo "Usage: $0 path/to/your_program.c"
    exit 1
fi

command -v valgrind &>/dev/null || { echo "This script requires valgrind, Aborting."; exit 1; }
[ $(uname) = Linux ] || echo "this bash script opens a pipe with O_RDWR, this works in Linux but might hang on other systems. Be warned."

make || { echo "SEEMINGLY INCOMPATIBLE"; exit; }

valgrind_grep(){
    grep 'lost: .* [1-9]' | tr -s ' ' | cut -d ' ' -f 2- | awk '{print "    " $0}'
}

# check if the shipped example file works well
echo -n "test.c output:   "
[[ "$(timeout -k 0.1s 1s ./test)" == $'FirstMessage\nHello world!' ]] && echo ok || echo FAIL

X=$(mktemp -u)
mkfifo "$X"
exec 3<>"$X"
rm "$X"

# all else
run() {
    OUTPUT=$(
        timeout -k 0.1s 2s valgrind --error-exitcode=13 "$@" 2> >(valgrind_grep>&3)
        RV=$?
        case "$RV" in
            124) echo TIMEOUT;;
            137) echo TIMEOUT+KILL;;
            13) echo MEMCHECK;;
            0) ;;
            *) echo FAIL "($( ((RV>128 && RV<=192)) && kill -l $((RV-128)) || echo $RV ))";;
        esac
    )
    echo "$OUTPUT"
    while read -t 0.01 <&3
    do
        echo "                   $REPLY" >&2
    done
}

echo "valgrind test.c: "
run ./test >/dev/null

run ./01_ids_make
run ./02_send_receive
run ./03_queue_size1
run ./04_queue_size2
run ./05_send_cancel
run ./06_send_badid
run ./07_mcast
run ./08_gather_basic
run ./09_gather_many
run ./10_multiples
run ./11_order
