#pragma once
#include <stdatomic.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>
#include <errno.h>
#include <threads.h>
#include <time.h>

// writing debug message to file descriptor 3 for the purpose of checking with check.sh
// to see the messages while running the programs in bash, add the redirection '3>&1'
#define assertVerbose(COND, STR, ...) if(!(COND)){ dprintf(3, STR "\n" __VA_OPT__(,) __VA_ARGS__); assert((COND)); }

/**
 * NEVER PUT VARIABLES IN A HEADER FILE UNLESS YOU REALLY KNOW WHAT YOU ARE DOING
 * (or sooner or later you'll have to deal with 'xyz defined multiple times' error)
 */

pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cv = PTHREAD_COND_INITIALIZER;
int state = 0;

void adv_state(int i){
    pthread_mutex_lock(&m);
    assert(state==i-1);
    state=i;
    pthread_cond_broadcast(&cv);
    pthread_mutex_unlock(&m);
}

void wait_state(int i){
    pthread_mutex_lock(&m);
    while(state<i)
        pthread_cond_wait(&cv, &m);
    pthread_mutex_unlock(&m);
}

struct thread_msg{
    atomic_int uc;
#if FLAVOR==1
    int num;
    double * dummy;
#else
    char * txt;
#endif
};

#if FLAVOR==1
thread_msg* thread_msg_make(int num){
    thread_msg* msg = malloc(sizeof(thread_msg));
    msg->dummy = malloc(2*sizeof(double));
    msg->dummy[0]=0x2.ap4; 
    msg->dummy[1]=0xdead.beefp00;
    msg->num=num;
    msg->uc=1;
    return msg;
}
#else
thread_msg* thread_msg_make(const char * str){
    thread_msg* msg = malloc(sizeof(thread_msg));
    if(str==NULL)
        msg->txt = NULL;
    else {
        msg->txt = malloc(strlen(str)+1);
        strcpy(msg->txt, str);
    }
    msg->uc=1;
    return msg;
}
#endif
 
 
void thread_msg_inc_use_count(thread_msg* msg){
    msg->uc++;
}
 
void thread_msg_dec_use_count(thread_msg* msg){
    if(!--msg->uc){
#if FLAVOR==1
        free(msg->dummy);
#else
        free(msg->txt);
#endif
        free(msg);
    }
}

thread_local struct timespec prev, now;
long elapsed(){
    clock_gettime(CLOCK_MONOTONIC, &now);
    long ms = (now.tv_sec-prev.tv_sec)*1000000l+(now.tv_nsec-prev.tv_nsec)/1000l;
    prev=now;
    return ms;
}
