#include "header.h"
#include <stdatomic.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
 
struct thread_msg{
    atomic_int uc;
    char * txt;
};
 
thread_msg* thread_msg_make(const char * str);
 
void t1(void*_);
 
void t2(void* arg);
 
int main(){
    thread_init();
 
    // create two more threads
    int id1 = thread_make(t1, NULL);
    int id2 = thread_make(t2, (void*)(intptr_t)id1);
 
    // send two messages to the first thread 
    thread_msg* msg1 = thread_msg_make("FirstMessage");
    thread_send(id1, msg1);
    // the second message contains the ID of the second thread
    thread_msg* msg2 = thread_msg_make("X");
    sprintf(msg2->txt, "%d", id2);
    thread_send(id1, msg2);
 
    // send a message to the second thread
    thread_msg* msg3 = thread_msg_make("Hello ");
    thread_send(id2, msg3);
 
    // receive a message from the second thread
    int sender = id2;
    thread_msg* msg4;
    thread_gather(&sender, 1, &msg4);
    thread_msg_dec_use_count(msg4);
 
    usleep(50000);
    return 0;
}
 
void t1(void*_){
    // receive a message and print its contents
    int sender;
    thread_msg * msg;
    thread_receive(&sender, &msg);
    printf("%s\n", msg->txt);
    thread_msg_dec_use_count(msg);
 
    // receive a message and extract a number (the ID of the second thread) from it
    thread_receive(&sender, &msg);
    int id2;
    sscanf(msg->txt, "%d", &id2);
    thread_msg_dec_use_count(msg);
 
    // send a message to two threads
    int tgt[]={0, id2};
    msg = thread_msg_make("world!");
    thread_mcast(tgt, 2, msg);
    // notice that thread 0 does not retreive the message from the queue
}
 
void t2(void* arg){
    // receive messages from the two threads and output them
    int senders[]={0, (intptr_t) arg};
    thread_msg* msgs[2];
    thread_gather(senders, 2, msgs);
    printf("%s%s\n", msgs[0]->txt, msgs[1]->txt);
    thread_msg_dec_use_count(msgs[0]);
    thread_msg_dec_use_count(msgs[1]);
 
    // send a message to the main thread
    thread_msg * msg = thread_msg_make(NULL);
    thread_send(0, msg);
 
    thread_term();
}
 
thread_msg* thread_msg_make(const char * str){
    thread_msg* msg = malloc(sizeof(thread_msg));
    if(str==NULL)
        msg->txt = NULL;
    else {
        msg->txt = malloc(strlen(str)+1);
        strcpy(msg->txt, str);
    }
    msg->uc=1;
    return msg;
}
 
void thread_msg_inc_use_count(thread_msg* msg){
    msg->uc++;
}
 
void thread_msg_dec_use_count(thread_msg* msg){
    if(!--msg->uc){
        free(msg->txt);
        free(msg);
    }
}

